/*!
 * socktailf.c
 *
 * Copyright (c) 2015, the socktailf authors
 *
 *  Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#define _POSIX_C_SOURCE 200809L

#define SOCKTAILF_RELEASE_VERSION "1.1.0"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <pthread.h>

#define SOCKET_MAX_BACKLOGGED_CONNECTIONS 5

#define WARNING(str) fputs("socktailf: " str "\n", stderr)
#define FATAL(str) { fputs("socktailf: " str "\n", stderr); exit(EXIT_FAILURE); } 

static char            tempdir[]    = "/tmp/socktailf.XXXXXX";
static char            socketfile[] = "                     /socket";
static pthread_mutex_t mtxbuff;
static char**          lines        = NULL;
static size_t*         lens         = NULL;
static unsigned int    numlines     = 15;
static unsigned int    linepos      = 0;
static unsigned int    exit_after_stdin = 0;
static unsigned int    dump_on_exit = 0;
static unsigned int    display_buffer_path = 1;

void write_buffer(int fd);

/* DIRECTORY HANDLING */  

void init_and_display_tempdir() {
	
	if ( mkdtemp(tempdir) == NULL ) {
		FATAL("Failed to create temporary directory");
	}

	memcpy(socketfile, tempdir, sizeof(tempdir) - 1);
	
	if (display_buffer_path) {
		fputs(socketfile, stdout);
		fputc('\n', stdout);
		fflush(stdout);
	}
	
}

int cleanup_and_exit(int exit_status) {
	
	if (dump_on_exit) {
		write_buffer(STDOUT_FILENO);
	}
	
	if ( unlink(socketfile) == -1 || rmdir(tempdir) == -1 ) {
		WARNING("Couldn't clean up");
	}
	
	exit(exit_status);
	return exit_status; 
}

/* SIGNAL HANDLING */

void handle_sig(int dummy) {
	cleanup_and_exit(EXIT_FAILURE);
}

void init_signalhandlers() {
	signal(SIGINT,  handle_sig);
	signal(SIGTERM, handle_sig);
}

/* APPLICATION THREADS */

void init_readbuffer() {
	assert(lines == NULL);
	
	lines = calloc(numlines, sizeof(char*));
	assert(lines != NULL);
	
	lens = calloc(numlines, sizeof(size_t));
	
	linepos = 0;
}

void* reading_thread(void* unused) {
	
	size_t  len  = 0;
	char*   line = NULL;
	ssize_t read = -1;
	
	for(;;) {
		
		line = NULL;
		len  = 0;
		read = getline(&line, &len, stdin);
		
		if (read == -1) {
			break; 
		}
		
		pthread_mutex_lock( &mtxbuff ); {
			
			if (lines[linepos] != NULL) {
				free(lines[linepos]);
			}
			lines[linepos] = line;
			lens[linepos]  = read;
			
			linepos = (linepos + 1) % numlines;
			
		} pthread_mutex_unlock( &mtxbuff );
	}
	
	if (exit_after_stdin) {
		
		/* FIXME: hack around race condition if the other thread     */
		/*  hasn't yet reached accept();                             */
		
		sleep(1);
		raise(SIGTERM);
	}
	
	return NULL;
}

void write_buffer(int fd) {
	
	pthread_mutex_lock( &mtxbuff ); {
		
		unsigned int pos = linepos;
		for(;;) {
			
			if (lines[pos] != NULL) {
				write(fd, lines[pos], lens[pos]);
			}
			
			pos = (pos + 1) % numlines;
			
			if (pos == linepos) {
				break; /* reached start */
			}
			
		}
		
	} pthread_mutex_unlock( &mtxbuff );
}

void* listening_thread(void* unused) {
	
	// Create socket
	
	int server_fd = socket(AF_UNIX, SOCK_STREAM, 0);
	
	// Bind and listen
	
	struct sockaddr_un server_address = {0};
	server_address.sun_family = AF_UNIX;
	strncpy(server_address.sun_path, socketfile, sizeof(socketfile));
	
	if ( bind(server_fd, (struct sockaddr*)&server_address, sizeof(server_address)) != 0 ) {
		FATAL("bind() failure");
	}
	
	if ( listen(server_fd, SOCKET_MAX_BACKLOGGED_CONNECTIONS) != 0 ) {
		FATAL("listen() failure");
	}
	
	for(;;) {
		
		// Receive connection
		
		int client_fd = accept(server_fd, NULL, NULL);
		
		if (errno == EINTR) {
			break;
		}
		
		write_buffer(client_fd);
		
		close(client_fd);
		
	}
	
	return NULL;
}

/* ENTRY POINT */

void run_threads() {
	pthread_t tlisten = {0};
	pthread_t tread   = {0};

	pthread_create(&tlisten, NULL, listening_thread, NULL);
	pthread_create(&tread,   NULL, reading_thread,   NULL);
	
	pthread_join(tlisten, (void**)NULL);
	pthread_join(tread,   (void**)NULL);
	
}

void print_usage() {
	fputs(
"Usage:\n"
"  socktailf [options]\n"
"\n"
"Options:\n"
"  -d                         Write the current buffer to stdout upon exit\n"
"  --help                     Display help text\n"
"  -n, --lines %count         Set the number of lines to buffer (default 15)\n"
"  -q                         Supress display of path to socket file\n"
"  -v, --version              Display version number and exit\n"
"  -x                         Exit after failure to read from stdin (e.g. EOF)\n"
		,
		stdout
	);
}

int main(int argc, char** argv) {
	
	/* process arguments */
	
	#define ARG_IS(x) (!strcmp(argv[i], x))
	
	for (int i = 1; i < argc; ++i) {
		
		int args_remaining = argc - i - 1;
		
		if ( args_remaining >= 1 && (ARG_IS("-n") || ARG_IS("--lines"))) {
			int param = atoi(argv[++i]);
			if (param < 1) {
				FATAL("Number of lines must be at least one");
			}
			numlines = param;
			
		} else if (ARG_IS("-d")) {
			dump_on_exit = 1;
			
		} else if (ARG_IS("-x")) {
			exit_after_stdin = 1;
			
		} else if (ARG_IS("-q")) {
			display_buffer_path = 0;
			
		} else if (ARG_IS("-v") || ARG_IS("--version")) {
			fputs( SOCKTAILF_RELEASE_VERSION "\n" , stdout );
			return EXIT_SUCCESS;			
			
		} else if (ARG_IS("--help")) {
			print_usage();
			return EXIT_SUCCESS;
			
		} else {
			fprintf(stderr, "Invalid argument '%s', try --help\n", argv[i]);
			return EXIT_FAILURE;
			
		}
		
	}
	
	/* run application */
	
	if ( pthread_mutex_init(&mtxbuff, NULL) != 0 ) {
		FATAL("pthread_mutex_init() failure");
	}
	
	init_readbuffer();	
	init_and_display_tempdir();	
	init_signalhandlers();
	
	run_threads();
	
	return cleanup_and_exit(EXIT_SUCCESS);
}
